from .language import Language
from .location import Location, LocationRange, FileLocationRange, FileLine
from .replacement import Replacement
from .transformation import Transformation
from .operator import Operator
from .mutation import Mutation
from .mutant import Mutant
from .constraint import Constraint



__all__ = ['Replacement']

from .location import FileLocationRange


# FIXME: use attrs
class Replacement(object):
    """
    Describes the replacement of a contiguous body of text in a single source
    code file with a provided text.
    """
    def __init__(self,
                 location: FileLocationRange,
                 text: str
                 ) -> None:
        self.__location = location
        self.__text = text

    def __repr__(self) -> str:
        return "Replacement({}, {})".format(repr(self.location),
                                            self.text)

    @property
    def filename(self) -> str:
        """
        The name of the file in which the replacement should be made.
        """
        return self.__location.filename

    @property
    def location(self) -> FileLocationRange:
        """
        The contiguous range of text that should be replaced.
        """
        return self.__location

    @property
    def text(self) -> str:
        """
        The source text that should be used as a replacement.
        """
        return self.__text



from typing import Iterator, Any, List, Optional, Dict
import logging

from .languages import Languages
from ..core import Operator
from ..exceptions import LanguageNotFound

logger = logging.getLogger(__name__)

__all__ = ['Operators']


class Operators(object):
    """
    Maintains information about the mutation operators that are supported
    by Boggart.
    """
    @staticmethod
    def from_defs(defs: List[Any],
                  languages: 'Languages',
                  base: 'Operators' = None
                  ) -> 'Operators':
        """
        Loads an operator configuration from a list of definitions taken
        from a configuration file, together with an optionally provided
        parent (overall) configuration.

        Raises:
            LanguageNotFound: if one of the operators given by the provided
              definitions supports a language that is not contained within
              the given configuration.
        """
        logger.info("Loading operators from definitions: %s", defs)
        operators = base if base else Operators()
        for d in defs:
            logger.info("Loading operator from definition: %s", d)
            op = Operator.from_dict(d)
            operators = operators.add(op, languages)
        logger.info("Loaded operators from definitions.")
        return operators

    def __init__(self,
                 operators: Optional[Dict[str, Operator]] = None
                 ) -> None:
        """
        Constructs a collection of operators for a set of operators provided
        in the form a dictionary, indexed by name.
        """
        self.__operators = dict(operators) if operators else {}

    def add(self, op: Operator, languages: Languages) -> 'Operators':
        """
        Returns a variant of this collection of mutation operators that also
        includes a given mutation operator.

        Params:
            op: the mutation operator that should be added.
            languages: the collection of languages that are supported by a
              particular configuration of Boggart.

        Raises:
            LanguageNotFound: if the given mutation operator supports a
              language that is not contained within the provided collection
              of languages.
        """
        logger.info("Adding operator to collection: %s", op)
        for supported_language in op.languages:
            if supported_language not in languages:
                logger.error("Failed to add operator, %s, to collection: language not found (%s)",  # noqa: pycodestyle
                             op, supported_language)
                raise LanguageNotFound(supported_language)

        ops = dict(self.__operators)
        ops[op.name] = op
        logger.info("Added operator to collection: %s", op)
        return Operators(ops)

    def __len__(self) -> int:
        """
        Returns the number of operators contained within this collection.
        """
        return len(self.__operators)

    def __iter__(self) -> Iterator[Operator]:
        """
        An iterator over the operators contained within this collection.
        """
        return self.__operators.values().__iter__()

    def __getitem__(self, name: str) -> Operator:
        """
        Attempts to fetch the definition of the mutation operator associated
        with a given name.

        Raises:
            KeyError: if no mutation operator is found with the given name.
        """
        return self.__operators[name]

    def __contains__(self, name: str) -> bool:
        """
        Determines whether this collection of mutation operators contains one
        with a given name.
        """
        return name in self.__operators



from typing import Iterator, Dict

from .api import API
from ..core import Language

__all__ = ['LanguageCollection']


class LanguageCollection(object):
    """
    Provides read-only access to the set of languages supported by a given
    server.
    """
    def __init__(self, api: API) -> None:
        """
        Constructs a new language collection for a given server.

        Parameters:
            api: the low-level client API that should be used to interact
                with the server.
        """
        language_dict_list = api.get("/languages").json()
        languages = [Language.from_dict(d) for d in language_dict_list]
        self.__contents = \
            {lang.name: lang for lang in languages}  # type: Dict[str, Language]  # noqa: pycodestyle

    def __iter__(self) -> Iterator[Language]:
        """
        Returns an iterator over the languages within this collection.
        """
        return self.__contents.values().__iter__()

    def __len__(self) -> int:
        """
        Returns a count of the number of languages within this collection.
        """
        return len(self.__contents)

    def __getitem__(self, name: str) -> Language:
        """
        Returns the language associated with a given name.

        Raises:
            KeyError: if no language is found with the given name.
        """
        return self.__contents[name]



__all__ = ['MutantCollection']

from typing import Iterator
from uuid import UUID

from .api import API
from ..core import Mutant


class MutantCollection(object):
    """
    Provides access to the mutants that are registered with a given server.
    """
    def __init__(self, api: API) -> None:
        """
        Constructs a new mutant collection for a given server.

        Parameters:
            api: the low-level client API that should be used to interact
                with the server.
        """
        self.__api = api

    def clear(self) -> None:
        """
        Destroys all registered mutants.
        """
        r = self.__api.delete("/mutants")
        if r.status_code != 204:
            self.__api.handle_erroneous_response(r)

    def __iter__(self) -> Iterator[UUID]:
        """
        Returns an iterator over the UUIDs of mutants within this collection.
        """
        uuid_str_list = self.__api.get("/mutants").json()
        uuid_list = [UUID(hex=s) for s in uuid_str_list]
        yield from uuid_list

    def __len__(self) -> int:
        """
        Returns a count of the number of mutants in this collection.
        """
        return len(list(self))

    def __delitem__(self, uuid: UUID) -> None:
        """
        Destroys the mutant associated with a given UUID.

        Raises:
            KeyError: if no mutant is found with the given UUID.
        """
        r = self.__api.delete('mutants/{}'.format(uuid.hex))
        if r.status_code == 204:
            return
        if r.status_code == 404:
            msg = "no mutant found with given UUID: {}"
            msg = msg.format(uuid.hex)
            raise KeyError(msg)
        self.__api.handle_erroneous_response(r)

    def __getitem__(self, uuid: UUID) -> Mutant:
        """
        Returns the mutant associated with a given UUID.

        Raises:
            KeyError: if no mutant is found with the given UUID.
        """
        r = self.__api.get('mutants/{}'.format(uuid.hex))
        if r.status_code == 200:
            return Mutant.from_dict(r.json())
        if r.status_code == 404:
            msg = "no mutant found with given UUID: {}"
            msg = msg.format(uuid.hex)
            raise KeyError(msg)
        self.__api.handle_erroneous_response(r)



__all__ = ['Constraint', 'IsSingleTerm', 'PrecededBy']

from typing import Any, Dict, FrozenSet
import attr
import logging

from rooibos import Match
from bugzoo import Bug as Snapshot

logger = logging.getLogger(__name__)  # type: logging.Logger


@attr.s(frozen=True)
class Constraint(object):
    @staticmethod
    def from_dict(d: dict) -> 'Constraint':
        if d['type'] == 'is-single-term':
            return IsSingleTerm.from_dict(d)
        if d['type'] == 'preceded-by':
            return PrecededBy.from_dict(d)
        raise SyntaxError("unrecognized constraint")

    def is_satisfied_by(self,
                        match: Match,
                        content_file: str,
                        offset_start: int,
                        offset_stop: int
                        ) -> bool:
        """
        Checks whether a given match satisfies this constraint.
        """
        raise NotImplementedError

    def to_dict(self) -> Dict[str, Any]:
        raise NotImplementedError


@attr.s(frozen=True)
class IsSingleTerm(Constraint):
    @staticmethod
    def from_dict(d: dict) -> 'IsSingleTerm':
        assert 'hole' in d
        hole = str(d['hole'])
        return IsSingleTerm(hole)

    hole = attr.ib(type=str)

    def is_satisfied_by(self,
                        match: Match,
                        content_file: str,
                        offset_start: int,
                        offset_stop: int
                        ) -> bool:
        # NOTE bug in rooibos causes empty terms to be omitted
        if self.hole not in match.environment:
            return False

        content = match.environment[self.hole].fragment.strip()
        return ' ' not in content

    def to_dict(self) -> Dict[str, Any]:
        return {'type': 'is-single-term',
                'hole': self.hole}


@attr.s(frozen=True)
class PrecededBy(Constraint):
    @staticmethod
    def from_dict(d: dict) -> 'PrecededBy':
        assert 'any-of' in d
        assert isinstance(d['any-of'], list)
        assert d['any-of'] != []
        assert all(isinstance(k, str) for k in d['any-of'])
        return PrecededBy(d['any-of'])

    options = attr.ib(type=FrozenSet[str],
                      converter=frozenset)  # type: ignore

    def is_satisfied_by(self,
                        match: Match,
                        content_file: str,
                        offset_start: int,
                        offset_stop: int
                        ) -> bool:
        preceded_by = content_file[:offset_start].rstrip()
        return any(preceded_by.endswith(opt) for opt in self.options)

    def to_dict(self) -> Dict[str, Any]:
        return {'type': 'preceded-by',
                'any-of': list(self.options)}



from typing import Dict, Any

from .location import FileLocationRange
from ..exceptions import BadFormat

__all__ = ['Mutation']


class Mutation(object):
    """
    Describes a concrete application of a given mutation operator at a specific
    location (range) in a source code file.
    """
    @staticmethod
    def from_dict(d: dict) -> 'Mutation':
        """
        Constructs a mutation from a dictionary-based description.

        Raises:
            BadFormat: if the provided dictionary does not match the expected
                format.
        """
        if not isinstance(d, dict):
            raise BadFormat("expected data structure to be a dictionary.")  # noqa: pycodestyle
        if 'operator' not in d:
            raise BadFormat("expected 'operator' property.")
        if 'transformation-index' not in d:
            raise BadFormat("expected 'transformation-index' property.")
        if 'location' not in d:
            raise BadFormat("expected 'location' property.")
        if 'arguments' not in d:
            raise BadFormat("expected 'arguments' property.")

        operator = d['operator']
        transformation_index = d['transformation-index']
        arguments = d['arguments']
        try:
            location = FileLocationRange.from_string(d['location'])
        except Exception:
            raise BadFormat("failed to parse location string.")

        return Mutation(operator, transformation_index, location, arguments)

    def __init__(self,
                 operator: str,
                 transformation_index: int,
                 at: FileLocationRange,
                 args: Dict[str, str]
                 ) -> None:
        self.__operator = operator
        self.__transformation_index = transformation_index
        self.__at = at
        self.__args = dict(args)

    def __repr__(self) -> str:
        return "Mutation({}, {}, {}, {})".format(self.operator,
                                                 self.transformation_index,
                                                 repr(self.at),
                                                 repr(self.arguments))

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Mutation) and \
               self.operator == other.operator and \
               self.transformation_index == other.transformation_index and \
               self.location == other.location and \
               self.arguments == other.arguments

    @property
    def operator(self) -> str:
        """
        The name of the mutation operator that was applied.
        """
        return self.__operator

    @property
    def transformation_index(self) -> int:
        """
        The index of the transformation that was applied.
        """
        return self.__transformation_index

    @property
    def location(self) -> FileLocationRange:
        """
        The character range at which the operator was applied.
        """
        return self.__at

    at = location

    @property
    def arguments(self) -> Dict[str, str]:
        """
        A dictionary of named arguments that are supplied to the operator for
        this mutation.
        """
        return self.__args.copy()

    def to_dict(self) -> dict:
        """
        Provides a dictionary-based description of this mutation, ready to be
        serialized.
        """
        return {
            'operator': self.operator,
            'transformation-index': self.transformation_index,
            'location': str(self.location),
            'arguments': self.arguments
        }



import os
import glob
import setuptools

path = os.path.join(os.path.dirname(__file__), 'src/boggart/version.py')
with open(path, 'r') as f:
    exec(f.read())

setuptools.setup(
    name='boggart',
    version=__version__,
    description='Lightweight, extensible, language-independent mutation testing.',
    long_description='TBA',
    author='Chris Timperley',
    author_email='christimperley@gmail.com',
    url='https://github.com/squaresLab/Hulk',
    license='mit',
    python_requires='>=3.5',
    install_requires=[
        'bugzoo>=2.1.11',
        'rooibos>=0.3.0',
        'attrs>=17.2.0',
        'pyyaml',
        'requests>=2.0.0',
        'flask>=0.10',
        'Flask-API'
    ],
    setup_requires=[
        'pytest-runner'
    ],
    tests_require=[
        'pytest'
    ],
    include_package_data=True,
    packages=setuptools.find_packages('src'),
    package_dir={'': 'src'},
    package_data={
        '': ['*.yml']
    },
    py_modules=[
        splitext(basename(path))[0] for path in glob.glob('src/*.py')
    ],
    entry_points = {
        'console_scripts': [ 'boggartd = boggart.server:main' ]
    },
    test_suite = 'tests'
)



from typing import Optional, Union, Dict, List, Iterator, Tuple
import logging

from bugzoo.core.patch import Patch
from bugzoo.core.bug import Bug
from bugzoo.util import indent

from .api import API
from .languages import LanguageCollection
from .operators import OperatorCollection
from .mutants import MutantCollection
from ..exceptions import *
from ..core import Operator, Language, Mutation, Mutant


logger = logging.getLogger(__name__)


__all__ = ['Client']


class Client(object):
    """
    A client for communicating with a boggart server.
    """
    def __init__(self,
                 base_url: str,
                 *,
                 timeout: int = 30,
                 timeout_connection: int = 60
                 ) -> None:
        """
        Constructs a new client for communicating with a given boggart server.

        Parameters:
            base_url:   the URL of the boggart server.
            timeout:    the default timeout for API calls (in seconds).

        Raises:
            ValueError: if the provided URL lacks a scheme (e.g., 'http').
        """
        logger.info("constructing client for boggart server: %s", base_url)
        self.__api = API(base_url,
                         timeout=timeout,
                         timeout_connection=timeout_connection)
        self.__languages = LanguageCollection(api=self.api)
        self.__operators = OperatorCollection(api=self.api)
        self.__mutants = MutantCollection(api=self.api)
        logger.info("registered languages: %s",
                    ', '.join([l.name for l in self.__languages]))
        logger.info("registered operators: %s",
                    ', '.join([op.name for op in self.__operators]))
        logger.info("constructed client for boggart server: %s", base_url)

    @property
    def base_url_bugzoo(self) -> str:
        """
        The base URL of the BugZoo server to which this boggart server is
        attached.
        """
        raise NotImplementedError

    @property
    def base_url_rooibos(self) -> str:
        """
        The base URL of the Rooibos server to which this boggart server is
        attached.
        """
        raise NotImplementedError

    @property
    def api(self) -> API:
        """
        The low-level client API used to communicate with the server.
        """
        return self.__api

    @property
    def languages(self) -> LanguageCollection:
        """
        The set of languages that are supported (i.e., can be mutated)
        by the server.
        """
        return self.__languages

    @property
    def mutants(self) -> MutantCollection:
        """
        The set of mutants that are registered with the server.
        """
        return self.__mutants

    @property
    def operators(self) -> OperatorCollection:
        """
        The set of mutation operators that are supported by the server.
        """
        return self.__operators

    def mutations_to_diff(self,
                          snapshot: Bug,
                          mutations: List[Mutation]
                          ) -> Patch:
        """
        Transforms a given set of mutations to a snapshot into a unified diff.
        """
        logger.info("transforming mutations to snapshot [%s] into a diff",  # noqa: pycodestyle
                    snapshot.name)
        path = "diff/mutations/{}".format(snapshot.name)
        payload = {
            'mutations': [m.to_dict() for m in mutations]
        }

        response = self.api.put(path, json=payload)
        if response.status_code == 200:
            diff = Patch.from_unidiff(response.text)
            diff_s = "[DIFF]\n{}\n[/DIFF]".format(indent(str(diff), 2))
            logger.info("transformed mutations to snapshot [%s] to diff:\n%s",
                        snapshot.name, indent(diff_s, 2))
            return diff
        else:
            logger.info("an error occurred whilst attempting to transform mutations to snapshot into a diff.")  # noqa: pycodestyle
            self.__api.handle_erroneous_response(response)

    def mutations(self,
                  snapshot: Bug,
                  filepath: str,
                  *,
                  language: Optional[Language] = None,
                  operators: Optional[List[Operator]] = None,
                  restrict_to_lines: Optional[List[int]] = None
                  ) -> Iterator[Mutation]:
        """
        Returns an iterator over all of the mutations that can be applied to
        a given file belonging to a BugZoo snapshot.

        Parameters:
            snapshot: the BugZoo snapshot.
            filepath: the path to the source code file inside the snapshot,
                relative to its source directory.
            language: the language that the source code file is written in. If
                left unspecified, boggart will attempt to automatically
                determine the language based on the file ending.
            operators: an optional list of mutation operators that should be
                used to generate mutations. If no list is provided, then all
                registered mutation operators for the specified language will
                be used as a default.

        Returns:
            an iterator over the possible mutations.

        Raises:
            SnapshotNotFound: if the given snapshot does not appear to be
                registered with the BugZoo server that is attached to this
                boggart server.
            FileNotFound: if no file is found with the given name in the
                snapshot.
        """
        assert operators is None or len(operators) > 0

        logger.info("Finding mutations in file '%s' belonging to snapshot '%s'.",  # noqa: pycodestyle
                    filepath,
                    snapshot.name)
        path = "mutations/{}/{}".format(snapshot.name, filepath)
        params = {}
        if language:
            params['language'] = language.name
        if operators:
            params['operators'] = ';'.join([op.name for op in operators])

        response = self.api.get(path, params)

        if response.status_code == 200:
            for jsn_mutation in response.json():
                logger.debug("Decoding mutation.",
                             extra={'mutation': jsn_mutation})
                mutation = Mutation.from_dict(jsn_mutation)
                logger.info("Found mutation: %s", repr(mutation))
                yield mutation
        else:
            logger.info("An error occurred whilst attempting to find mutations.")  # noqa: pycodestyle
            self.__api.handle_erroneous_response(response)

    def mutate(self,
               snapshot: Bug,
               mutations: List[Mutation]
               ) -> Mutant:
        """
        Applies a given mutation to a snapshot.

        Parameters:
            snapshot: the snapshot that should be mutated.
            mutations: the mutations to apply to the snapshot.

        Returns:
            a description of the generated mutant.
        """
        payload = {
            'snapshot': snapshot.name,
            'mutations': [m.to_dict() for m in mutations]
        }
        logger.info("Applying mutations to snapshot '%s': %s",
                    snapshot.name,
                    ', '.join(repr(m) for m in mutations),
                    extra=payload)
        response = self.api.post("mutants", json=payload)

        if response.status_code == 200:
            mutant = Mutant.from_dict(response.json())
            logger.info("Applied mutations to snapshot '%s' to generate mutant: %s",  # noqa: pycodestyle
                        snapshot.name,
                        repr(mutant),
                        extra={'mutant': response.json()})
            return mutant
        else:
            logger.info("An error occurred whilst attempting to mutate snapshot.")  # noqa: pycodestyle
            self.__api.handle_erroneous_response(response)



__all__ = ['Transformation']

from typing import Any, Dict, FrozenSet
import attr
import logging

from .constraint import Constraint

from rooibos import Match

logger = logging.getLogger(__name__)  # type: logging.Logger


@attr.s(frozen=True)
class Transformation(object):
    """
    Describes a source code transformation as a corresponding pair of Rooibos
    match and rewrite templates.
    """
    match = attr.ib(type=str)
    rewrite = attr.ib(type=str)
    constraints = attr.ib(type=FrozenSet[Constraint],
                          converter=frozenset)  # type: ignore

    @staticmethod
    def from_dict(d: dict) -> 'Transformation':
        assert 'match' in d
        assert 'rewrite' in d
        assert isinstance(d['match'], str)
        assert isinstance(d['rewrite'], str)

        match = d['match']
        rewrite = d['rewrite']
        constraints = \
            [Constraint.from_dict(c) for c in d.get('constraints', [])]

        return Transformation(match, rewrite, constraints)  # type: ignore

    def satisfies_constraints(self,
                              match: Match,
                              content_file: str,
                              offset_start: int,
                              offset_stop: int
                              ) -> bool:
        """
        Checks whether a given match satisfies the constraints of this
        transformation.
        """
        return all(c.is_satisfied_by(match,
                                     content_file,
                                     offset_start,
                                     offset_stop)
                   for c in self.constraints)

    def to_dict(self) -> dict:
        """
        Provides a dictionary--based description of this transformation, ready
        to be serialized.
        """
        return {'match': self.match,
                'rewrite': self.rewrite,
                'constraints': [c.to_dict() for c in self.constraints]}
